#pragma once

#if PLATFORM_WINDOWS
#include "Windows/AllowWindowsPlatformTypes.h"
#endif
#include <Windows.h>
#include <stdio.h>
#include <conio.h>
#include <d3d11.h>
#if PLATFORM_WINDOWS
#include "Windows/HideWindowsPlatformTypes.h"
#endif


#include "MixCastLog.h"

struct SharedSurfaceWriterMetadata
{
	unsigned long sharedHandle;

	unsigned long long syncTime;
	unsigned long long renderTime;

	unsigned long frameIndex;
};

class SharedSurfaceWriter
{
public:
	ID3D11Device* d3dDevice = NULL;
	ID3D11Texture2D* srcTex = NULL;

	ID3D11DeviceContext* fenceContext = NULL;
	ID3D11Query* fenceQuery = NULL;

	SharedSurfaceWriter(LPCWSTR frameId)
	{
		LPWSTR key = new TCHAR[256];
		//Events
		wsprintf(key, TEXT("Global\\MixCast.SharedTex(%s).ReadStart"), frameId);
		readyForReadSignal = CreateEvent(NULL, true, false, key);
		ResetEvent(readyForReadSignal);
		wsprintf(key, TEXT("Global\\MixCast.SharedTex(%s).WriteStart"), frameId);
		readyForWriteSignal = CreateEvent(NULL, true, false, key);
		ResetEvent(readyForWriteSignal);

		//Metadata
		wsprintf(key, TEXT("MixCast-SharedTex(%s)-Metadata"), frameId);
		DWORD metadataSize = sizeof(SharedSurfaceWriterMetadata);
		metadataFile = CreateFileMapping(
			INVALID_HANDLE_VALUE,		// use paging file
			NULL,						// default security
			PAGE_READWRITE,				// read/write access
			0,							// maximum object size (high-order DWORD)
			metadataSize,  // maximum object size (low-order DWORD)
			key);						// name of mapping object
		sharedMetadata = (SharedSurfaceWriterMetadata*)MapViewOfFile(metadataFile,   // handle to map object
			FILE_MAP_ALL_ACCESS, // read/write permission
			0,
			0,
			metadataSize);
		ZeroMemory(sharedMetadata, metadataSize);

		delete[] key;

		Update(NULL, 0, 0, 0);
	}
	~SharedSurfaceWriter()
	{
		ClearResources();

		UnmapViewOfFile(sharedMetadata);
		CloseHandle(metadataFile);

		CloseHandle(readyForReadSignal);
		CloseHandle(readyForWriteSignal);

		SAFE_RELEASE(sharedTex);
	}

	void CheckFenceStatus()
	{
		if (fenceQuery != NULL)
		{
			if (fenceContext->GetData(fenceQuery, NULL, 0, 0) == S_OK)
			{
				SAFE_RELEASE(fenceContext);
				SAFE_RELEASE(fenceQuery);
				SetEvent(readyForReadSignal);
			}
		}
	}
	void EnqueueFenceCheck()
	{
		ENQUEUE_RENDER_COMMAND(void)(
			[&](FRHICommandListImmediate& RHICmdList)
			{
				CheckFenceStatus();
			});
	}
	bool WaitForReadyToWrite(int timeout)
	{
		if (timeout == -1) {
			WaitForSingleObject(readyForWriteSignal, INFINITE);
			ResetEvent(readyForWriteSignal);
			return true;
		}
		else {
			DWORD res = WaitForSingleObject(readyForWriteSignal, timeout);
			if (res == WAIT_OBJECT_0)
			{
				ResetEvent(readyForWriteSignal);
				return true;
			}
			return false;
		}
	}
	void Update(ID3D11Texture2D* newSrcTex, unsigned long long newSyncTime, unsigned long long newRenderTime, unsigned long newFrameIndex)
	{
		if (srcTex != newSrcTex)
		{
			srcTex = newSrcTex;
			ENQUEUE_RENDER_COMMAND(void)(
				[&](FRHICommandListImmediate& RHICmdList)
			{
				ClearResources();
				CreateResources();

				sharedMetadata->sharedHandle = sharedHandle;
			});
		}

		sharedMetadata->syncTime = newSyncTime;
		sharedMetadata->renderTime = newRenderTime;
		sharedMetadata->frameIndex = newFrameIndex;
	}
	void UpdateSharedSurface()
	{
		if (sharedTex == NULL)
			return;

		D3D11_QUERY_DESC queryDesc;
		queryDesc.Query = D3D11_QUERY_EVENT;
		queryDesc.MiscFlags = 0;
		d3dDevice->CreateQuery(&queryDesc, &fenceQuery);

		ID3D11DeviceContext* ctx = NULL;
		d3dDevice->GetImmediateContext(&ctx);

		ctx->CopyResource(sharedTex, srcTex);

		ctx->End(fenceQuery);
		fenceContext = ctx;		//Don't release the context yet since I'm assuming the Fence check needs it
	}
	void EnqueueTextureCopy()
	{
		ENQUEUE_RENDER_COMMAND(void)(
			[&](FRHICommandListImmediate& RHICmdList)
			{
				UpdateSharedSurface();
			});
	}
	void SignalEndOfStream()
	{
		Update(NULL, 0, 0, 0);

		SAFE_RELEASE(fenceContext);
		SAFE_RELEASE(fenceQuery);

		SetEvent(readyForReadSignal);
	}

private:
	HANDLE readyForReadSignal = NULL;
	HANDLE readyForWriteSignal = NULL;

	HANDLE metadataFile = NULL;
	SharedSurfaceWriterMetadata* sharedMetadata = NULL;

	ID3D11Texture2D* sharedTex = NULL;
	unsigned long sharedHandle = 0;
	int texWidth, texHeight = 0;
	DXGI_FORMAT texFmt;

	bool CreateResources()
	{
		if (srcTex == NULL)
		{
			sharedHandle = 0;
			texWidth = 0;
			texHeight = 0;
			texFmt = DXGI_FORMAT_UNKNOWN;
			return true;
		}

		srcTex->GetDevice(&d3dDevice);

		D3D11_TEXTURE2D_DESC srcDesc;
		srcTex->GetDesc(&srcDesc);

		D3D11_TEXTURE2D_DESC newDesc;
		newDesc.Width = srcDesc.Width;
		newDesc.Height = srcDesc.Height;
		newDesc.MipLevels = 1;
		newDesc.ArraySize = 1;
		newDesc.Format = EnsureShareableFormat(srcDesc.Format);
		newDesc.Usage = D3D11_USAGE_DEFAULT;
		newDesc.SampleDesc.Count = 1;
		newDesc.SampleDesc.Quality = 0;
		newDesc.BindFlags = D3D11_BIND_SHADER_RESOURCE;
		newDesc.CPUAccessFlags = 0;
		newDesc.MiscFlags = D3D11_RESOURCE_MISC_SHARED;

		try
		{
			HRESULT hr = d3dDevice->CreateTexture2D(&newDesc, NULL, &sharedTex);

			if (FAILED(hr)) {
				return false;
			}
		}
		catch (...)
		{
			MIXCAST_ERROR("Failed to create shared texture");
			return false;
		}

		HANDLE sharedHandleVal;

		IDXGIResource* dxgi_res;
		try
		{
			HRESULT hr = sharedTex->QueryInterface(__uuidof(IDXGIResource), (void**)&dxgi_res);

			if (FAILED(hr)) {
				MIXCAST_ERROR("Failed to get dxgi resource");
				SAFE_RELEASE(sharedTex);
				return false;
			}

			dxgi_res->GetSharedHandle(&sharedHandleVal);
		}
		catch (...)
		{
			SAFE_RELEASE(dxgi_res);
			SAFE_RELEASE(sharedTex);

			MIXCAST_ERROR("Failed to get shared handle");
			return false;
		}


		SAFE_RELEASE(dxgi_res);

		sharedHandle = HandleToULong(sharedHandleVal);
		texWidth = newDesc.Width;
		texHeight = newDesc.Height;
		texFmt = newDesc.Format;
		return true;
	}

	void ClearResources()
	{
		if (sharedHandle == 0)
			return;

		//SAFE_RELEASE(d3dDevice);
		SAFE_RELEASE(sharedTex);

		//CloseHandle(ULongToHandle(sharedHandle));
		sharedHandle = 0;
	}

	DXGI_FORMAT EnsureShareableFormat(DXGI_FORMAT format)
	{
		if (format == DXGI_FORMAT_R8G8B8A8_UNORM_SRGB || format == DXGI_FORMAT_R8G8B8A8_TYPELESS)
		{
			return DXGI_FORMAT_R8G8B8A8_UNORM;
		}
		else if (format == DXGI_FORMAT_B8G8R8A8_UNORM_SRGB || format == DXGI_FORMAT_B8G8R8A8_TYPELESS)
		{
			return DXGI_FORMAT_B8G8R8A8_UNORM;
		}
		else if (format == DXGI_FORMAT_R32G32B32A32_TYPELESS)
		{
			return DXGI_FORMAT_R32G32B32A32_FLOAT;
		}
		else if (format == DXGI_FORMAT_R32_TYPELESS || format == DXGI_FORMAT_D32_FLOAT || format == DXGI_FORMAT_R32_FLOAT)
		{
			return DXGI_FORMAT_R32_FLOAT;
		}
		return format;
	}
};

